"""Constants"""

import curses

SPECIAL_KEYS: tuple[int] = tuple(
    (value for key, value in curses.__dict__.items() if key.startswith("KEY"))
)
KEY_ESC = 27
KEY_ENTER = 10
KEY_BACKSPACE = 127

CURSOR_HIDDEN = 0
CURSOR_VISIBLE = 1
CURSOR_BLOCK = 2



"""Component base class module"""

# pylint: disable=no-member,unused-import,unused-argument
from inspect import signature
import curses
from os import get_terminal_size, terminal_size
from typing import Callable, Self, TypeAlias, TypeGuard, TypeVar

from lymia.anim import Animator
from lymia.panel import Panel

from .utils import clear_line
from .data import ReturnType, status, SceneResult
from .menu import Menu
from .forms import Forms

Tcomp = TypeVar("Tcomp", bound="Scene")

DefaultCallback: TypeAlias = "Callable[[], ReturnType | SceneResult]"
WinCallback: TypeAlias = "Callable[[curses.window], ReturnType]"
DefaultMethod: TypeAlias = "Callable[[Tcomp], ReturnType | SceneResult]"
WinMethod: TypeAlias = "Callable[[Tcomp, curses.window], ReturnType]"

GenericFunction: TypeAlias = "DefaultCallback | WinCallback"
Method: TypeAlias = "DefaultMethod | WinMethod"
Function: TypeAlias = "GenericFunction | Method"


def uses_window(fn: Function) -> "TypeGuard[WinCallback | WinMethod]":
    """Check if a function uses a window param"""
    fn_signature = signature(fn)
    return "stdscr" in fn_signature.parameters


def is_method(fn: Function) -> TypeGuard[Method]:
    """Check if a function uses 'Self'"""
    fn_signature = signature(fn)
    return "self" in fn_signature.parameters

def is_method_and_uses_window(fn: Function) -> TypeGuard[WinMethod]:
    """Check if a function is a method and uses window param"""
    return uses_window(fn) and is_method(fn)

def no_op():
    """No op"""
    return ReturnType.CONTINUE


class SceneMeta(type):
    """Scene metaclass"""

    def __new__(mcs, name, bases, dct: dict):
        keymap = {}
        actions = {}
        for fn in dct.values():
            if hasattr(fn, "_keys"):
                for key in fn._keys:
                    key: int = ord(key) if isinstance(key, str) else key
                    keymap[key] = fn.__name__
                actions[fn.__name__] = fn
        dct["_keymap"] = keymap
        dct["_actions"] = actions
        return super().__new__(mcs, name, bases, dct)


class Scene(metaclass=SceneMeta):
    """Base class for all sorts of scenes"""

    generic_height: int = 3
    reserved_lines: int = 5
    should_clear: bool = True
    auto_resize: bool = True
    render_fps = -1
    minimal_size: tuple[int, int] = (-1, -1)

    _keymap: dict[int, str] = {}
    _actions: "dict[str, DefaultCallback | WinCallback]" = {}
    _last_menu_actions: list[str] = []


    def __init__(self) -> None:
        self._init = False
        self._override = False
        self._screen: curses.window = None # type: ignore
        self._panels: tuple[Panel, ...] = ()
        self._fps = 0
        self._animator: Animator

    def draw(self) -> None | ReturnType:
        """Draw this component"""
        raise NotImplementedError

    def keymap_override(self, key: int) -> ReturnType:
        """Override key component"""
        return ReturnType.REVERT_OVERRIDE

    def handle_key(self, key: int) -> "ReturnType | SceneResult":
        """Handle key component"""
        if self._override:
            ret = self.keymap_override(key)
            if ret == ReturnType.REVERT_OVERRIDE:
                self._override = False
                return ReturnType.CONTINUE
            return ret
        name = self._keymap.get(key, None)
        if not name:
            return ReturnType.CONTINUE

        action = self._actions.get(name, no_op)
        if action is no_op:
            return ReturnType.CONTINUE
        if is_method(action):
            if is_method_and_uses_window(action):
                return action(self, self._screen)
            return action(self) # type: ignore
        if uses_window(action):
            return action(self._screen) # type: ignore
        return action()  # type: ignore

    def show_status(self):
        """Show statuses"""
        height = self.height
        stdscr = self._screen
        try:
            # clear_line(stdscr, height - 1)
            stdscr.addstr(height - 1, 0, status.get())
        except curses.error: # occurs during resize
            pass

    def update_panels(self):
        """Update all panels"""
        for panel in self._panels:
            panel.draw()

    def syscall(self) -> ReturnType:
        """Do whatever you want."""
        return ReturnType.OK

    def leave(self):
        """Leave this component"""
        return ReturnType.BACK

    def init(self, stdscr: curses.window):
        """Initialize this component"""
        if self._init is True:
            return
        self._screen = stdscr
        self._init = True

    def on_unmount(self):
        """On unmount"""
        for panel in self._panels:
            panel.hide()
        return None

    def register_keymap(self, menu: Menu):
        """Register a menu's keymap into this component"""
        for action, (key, callback) in menu.get_keymap().items():
            self._keymap[key] = action
            self._actions[action] = callback

    def cleanup_menu_keymap(self):
        """Cleanup menu's register keymap"""
        for action in self._last_menu_actions:
            self._actions[action] = no_op
        self._last_menu_actions.clear()

    @property
    def term_size(self): # type: ignore
        """Return terminal size"""
        if self.auto_resize:
            return get_terminal_size()
        y, x = self.screen.getmaxyx()
        return terminal_size((x, y))

    @property
    def height(self):
        """Height of current terminal"""
        return self.term_size.lines

    @property
    def width(self):
        """Width of current terminal"""
        return self.term_size.columns

    @property
    def unreserved_lines(self):
        """Return unreserved lines"""
        return self.height - self.reserved_lines

    @property
    def screen(self):
        """This compoenent's painter"""
        if self._screen is None:
            raise RuntimeError("This component has not initialized yet")
        return self._screen

    @property
    def animator(self):
        """Component's animator"""
        if not hasattr(self, '_animator'):
            return None
        return self._animator

    @property
    def fps(self):
        """this Scene's FPS (not constant)"""
        return self._fps

    @fps.setter
    def fps(self, num: float):
        """this Scene's FPS (not constant)"""
        self._fps = num

    def __repr__(self) -> str:
        return f"<Component/{type(self).__name__}>"

class MenuFormScene(Scene):
    """Base for components with menu and forms"""

    def __init__(self, menu: Menu):
        super().__init__()
        self._menu = menu
        self._active_form: Forms | None = None
        self.register_keymap(menu)

    def draw(self):
        self._menu.draw(self._screen)
        if self._active_form:
            self._active_form.draw(self._screen)

    def keymap_override(self, key: int) -> ReturnType:
        if self._active_form:
            ret = self._active_form.handle_edit(key)
            if ret == ReturnType.REVERT_OVERRIDE:
                self._active_form = None
            return ret
        return ReturnType.REVERT_OVERRIDE

    def select_menu_item(self):
        """Select menu item"""
        _, item = self._menu.fetch()
        if isinstance(item, Forms):
            item()
            self._active_form = item
            self._override = True
            return ReturnType.OVERRIDE
        return ReturnType.CONTINUE

def on_key(*keys: str | int):
    """On key event binding"""

    def inner(fn: "Function"):
        fn._keys = keys  # pylint: disable=protected-access
        return fn

    return inner



"""Panels"""

# pylint: disable=no-member

import curses.panel
import curses
from typing import Any, Callable, ParamSpec


def refresh():
    """Refresh method"""
    curses.panel.update_panels()
    curses.doupdate()

Ps = ParamSpec("Ps")

REQ_REFRESH_METHODS = [
    "above",
    "below",
    "top",
    "bottom",
    "hide",
    "show",
    "move",
    "replace",
]


class Panel:
    """Panel"""

    __slots__ = ("_win", "_panel", "_draw")

    def __init__(
        self,
        height: int,
        width: int,
        start_x: int = 0,
        start_y: int = 0,
        callback: Callable[[curses.window, Any], None] | None = None,
        state: Any = None
    ) -> None:
        self._win = curses.newwin(height, width, start_x, start_y)
        self._panel = curses.panel.new_panel(self._win)
        self._panel.set_userptr(state)
        self._draw = callback

    def set_state(self, obj: Any):
        """Set panel's state"""
        self._panel.set_userptr(obj)

    def get_state(self):
        """Get panel's state"""
        return self._panel.userptr()

    @property
    def screen(self):
        """Return the screen this panel assosciates to"""
        return self._win

    @property
    def panel(self):
        """Return the panel this class references to"""
        return self._panel

    def draw(self):
        """Draw this panel"""
        if callable(self._draw):
            self._draw(self._win, self.get_state())

    def above(self):
        """Put panel above"""
        self._panel.above()
        refresh()

    def below(self):
        """Put panel below"""
        self._panel.below()
        refresh()

    def top(self):
        """Put panel on top"""
        self._panel.top()
        refresh()

    def bottom(self):
        """Put panel on bottom"""
        self._panel.bottom()
        refresh()

    def hide(self):
        """Hide"""
        self._panel.hide()
        refresh()

    def show(self):
        """Show"""
        self._panel.show()
        refresh()

    def move(self, x: int, y: int):
        """Move current panel"""
        try:
            self._panel.move(y, x)
        except curses.panel.error as exc:
            exc.add_note(f"x: {x} and y: {y}")
            raise exc
        refresh()

    def replace(self, win: curses.window):
        """Replaces current window"""
        self._panel.replace(win)
        self._win = win
        refresh()



"""Forms base class"""

# pylint: disable=no-name-in-module,no-member,unused-argument,too-many-instance-attributes

import curses
from typing import Any

from ..data import ReturnType
from .. import const


class Forms:
    """Base class for all sorts of forms"""

    LEFT_KEYS = (curses.KEY_LEFT,)
    RIGHT_KEYS = (curses.KEY_RIGHT,)

    def __init__(
        self,
        label: str,
        value: Any = None,
        suffix: str = ": ",
        margin_left: int = 0,
        field_pos: int = 0,
        style: int = 0,
    ) -> None:
        self._buffer = value
        self._label = label
        self._label_suffix = suffix
        self._margin_left = margin_left
        self._field_height_pos = field_pos
        self._style = style
        self._cursor = 0
        self._editing = False
        self._escdelay = curses.get_escdelay()
        self._prefix = ""

    @property
    def label(self):
        """Field label"""
        return self._label

    @property
    def editing(self):
        """If current field is in edit mode"""
        return self._editing

    @property
    def value(self):
        """Value of current field"""
        return self._buffer

    @property
    def displayed_value(self):
        """Displayed value of current field"""
        return self._buffer

    def display(self):
        """Returns pre-rendered of combination of label, suffix, and displayed value"""
        if not self._editing:
            return (
                f"{self._prefix}{self.label}{self._label_suffix}{self.displayed_value}"
            )
        return ""

    def __call__(self) -> None:
        self.enter_edit()

    def set_field_pos(self, num: int):
        """Set field position"""
        self._field_height_pos = num
        return self

    def set_prefix(self, prefix: str):
        """Set prefix"""
        self._prefix = prefix
        return self

    def reposition_cursor(self, stdscr: curses.window, prefix: int):
        """Reposition cursor"""
        stdscr.move(self._field_height_pos, prefix + self._cursor)

    def draw(self, stdscr: curses.window):
        """Draw this form"""
        content = (
            f"{self._prefix}{self._label}{self._label_suffix}{self.displayed_value}"
        )
        cursor_prefix = len(self._prefix) + len(self._label) + len(self._label_suffix)
        stdscr.addstr(self._field_height_pos, 0, content)
        if self._editing:
            self.reposition_cursor(stdscr, cursor_prefix)

    def enter_edit(self):
        """Enter edit mode"""
        curses.set_escdelay(1)
        curses.curs_set(2)
        self._editing = True
        self._cursor = len(self._buffer)

    def exit_edit(self):
        """Exit edit mode"""
        curses.set_escdelay(self._escdelay)
        curses.curs_set(0)
        self._editing = False

    def validate(self, char: str):
        """Validate current value"""
        return True

    def handle_edit(self, key: int):
        """Handle key edit"""
        if key in (curses.KEY_BACKSPACE, const.KEY_BACKSPACE):
            self._buffer = self._buffer[:-1]
            if self._cursor > 0:
                self._cursor -= 1
            return ReturnType.CONTINUE

        if key in self.LEFT_KEYS and self._cursor > 0:
            self._cursor -= 1
        if key in self.RIGHT_KEYS and self._cursor < len(self._buffer):
            self._cursor += 1

        if key in (curses.KEY_ENTER, const.KEY_ENTER):
            self.exit_edit()
            return ReturnType.REVERT_OVERRIDE

        if key == const.KEY_ESC:
            self._buffer = ""
            self.exit_edit()
            return ReturnType.REVERT_OVERRIDE

        if key in const.SPECIAL_KEYS:
            return ReturnType.CONTINUE
        char = chr(key)

        if not self.validate(char):
            return ReturnType.ERR

        if self._cursor != len(self._buffer):  # <-- Not at the end?
            self._buffer = (
                self._buffer[: self._cursor] + char + self._buffer[self._cursor :]
            )
            self._cursor += 1
            return ReturnType.CONTINUE
        self._buffer += char
        self._cursor += 1
        return ReturnType.CONTINUE

    def __repr__(self) -> str:
        return f"<Forms/{type(self).__name__}>"



from os import walk
from os.path import join

def catch():
    """catch some file"""
    known_files = []
    for root, _, files in walk('.'):
        for file in files:
            if file.endswith('.py'):
                known_files.append(join(root, file))
    return known_files

def readln(file: str):
    """Read lines in a file"""
    with open(file, encoding='utf-8') as f:
        return len(f.readlines())

def main():
    """main"""
    files = catch()
    total = 0
    for file in files:
        length = readln(file)
        total += length
        print(f"{file} -> has {length} lines")
    print(f"This project has {total} lines")

if __name__ == '__main__':
    main()



"""Data-related information"""
# pylint: disable=no-member,no-name-in-module

import curses
from enum import IntEnum
from typing import NamedTuple, Generic, TypeVar, TYPE_CHECKING

if TYPE_CHECKING:
    from .scene import Scene

T = TypeVar("T")

class ReturnType(IntEnum):
    """Return type"""

    OK = 0
    ERR = 1
    ERR_BACK = 2
    RETURN_TO_MAIN = 3
    BACK = 4
    CONTINUE = 5
    EXIT = -1

    OVERRIDE = 99
    REVERT_OVERRIDE = 100

class ReturnInfo(Generic[T], NamedTuple):
    """Return info"""
    type: ReturnType
    reason: str
    additional_info: T

class _StatusInfo:
    """Set status bar (bottom bar) information"""
    def __init__(self) -> None:
        self._data = " "

    def get(self) -> str:
        """Get current status info"""
        return self._data

    def set(self, value: str):
        """Set status info"""
        self._data = value

    def reset(self):
        """Reset status info"""
        self._data = " "

class SceneResult(NamedTuple):
    """Scene result, usable when returning Scene"""
    scene: "Scene"
    target: curses.window | None = None

status = _StatusInfo()



"""Screen"""

# pylint: disable=no-member,too-many-branches,pointless-string-statement
import curses
from typing import TypeAlias, overload

from lymia.errors import RenderError

"""
Fucking abomination, don't bother code this hell
"""


ChType: TypeAlias = str | bytes | int

def is_exclusively_different(a, b, type_):
    """Is a/b exclusively different?"""
    if isinstance(a, type_) and not isinstance(b, type_):
        return True
    if isinstance(b, type_) and not isinstance(a, type_):
        return True
    return False


class Screen:
    """High level screen impl"""

    def __init__(self, backend: curses.window) -> None:
        self._stdscr = backend

    @property
    def encoding(self):
        """Encoding of this screen"""
        return self._stdscr.encoding

    def _deduce_exc(self, y: int, x: int, obj_ref: str):
        content_length = len(obj_ref)
        maxy, maxx = self._stdscr.getmaxyx()
        if x + content_length >= maxx:
            return f"Content overflow for x: {x + content_length} > {maxx}"
        if y >= maxy:
            return f"Invalid height: {y} >= {maxy}"
        return "Unknown error"

    def _cook_err(self, main_msg: str, y: int, x: int, obj_ref: str, exc: curses.error):
        err = RenderError(main_msg)
        err.add_note(self._deduce_exc(y, x, obj_ref) or str(exc))
        return err

    @overload
    def add_char(self, ch: ChType, attr: int = 0):
        pass

    @overload
    def add_char(self, y: int, x: int, ch: ChType, attr: int = 0):
        pass

    def add_char(self, *args, **kwargs):
        """Paint character ch at (y, x) with attributes attr, overwriting any character previously
        painted at that location.
        By default, the character position and attributes are the
        current settings for the window object."""
        if len(args) == 1 or (len(args) == 2 and isinstance(args[1], int)):
            ch = args[0]
            attr = args[1] if len(args) == 2 else kwargs.get("attr", 0)
            x, y = kwargs.get("x", None), kwargs.get("y", None)
            if isinstance(x, int) and isinstance(y, int):
                try:
                    self._stdscr.addch(y, x, ch, attr)
                except curses.error as exc:
                    raise self._cook_err(
                        "Cannot render character", y, x, ch, exc
                    ) from None
                return
            if is_exclusively_different(x, y, int):
                raise TypeError("Inconsistent type for x and y")
            try:
                self._stdscr.addch(ch, attr)
            except curses.error as exc:
                posy, posx = self._stdscr.getyx()
                raise self._cook_err(
                    "Cannot render character", posy, posx, ch, exc
                ) from None

            return
        if len(args) >= 3:
            y, x, ch = args[:3]  # pylint: disable=unbalanced-tuple-unpacking
            attr = args[3] if len(args) > 3 else kwargs.get("attr", 0)
            try:
                self._stdscr.addch(y, x, ch, attr)
            except curses.error as exc:
                raise self._cook_err("Cannot render character", y, x, ch, exc) from None
            return
        ch = kwargs.get("ch", None)
        attr = kwargs.get("attr", 0)
        x, y = kwargs.get("x", None), kwargs.get("y", None)
        if ch is None and all((isinstance(x, int), isinstance(y, int))):
            raise ValueError("Cannot insert, passed char is None.")
        if ch is None:
            ch = args[0]  # type: ignore
        if is_exclusively_different(x, y, int):
            raise ValueError("Inconsistent type for x and y")
        if isinstance(x, int) and isinstance(y, int):
            try:
                self._stdscr.addch(x, y, ch, attr)
            except curses.error as exc:
                raise self._cook_err("Cannot render character", y, x, ch, exc) from None
            return
        try:
            self._stdscr.addch(ch, attr)
        except curses.error as exc:
            posy, posx = self._stdscr.getyx()
            raise self._cook_err(
                "Cannot render character", posy, posx, ch, exc
            ) from None



"""Menu (not top menu)"""

# pylint: disable=no-name-in-module,no-member

from collections.abc import Sequence
from math import inf
from os import get_terminal_size
import curses
from typing import Callable, Generic, TypeAlias, TypeVar

from lymia.colors import ColorPair
from lymia.forms import Forms

from .data import ReturnType
from .utils import prepare_windowed

T = TypeVar("T")
Fields: TypeAlias = (
    tuple[tuple[str, Callable[[], T]], ...] | tuple[tuple[str, Forms], ...]
)
FieldsFn: TypeAlias = Callable[[int], tuple[str, Callable[[], T]]]


class Menu(Generic[T]):
    """Menu helper component

    fields: Menu fields, consists of label/callback
    selected_style: Style for cursor
    margin: tuple[int, int], this margin is simplified as margin top and margin bottom
    max_height: Menu's maximum height

    For fields, callback can have a `.display()` that returns a pre-rendered label data
    The content returned by `.display()` is displayed as-is.
    """

    KEYMAP_UP = curses.KEY_UP
    KEYMAP_DOWN = curses.KEY_DOWN

    def __init__(
        self,
        fields: Fields | FieldsFn,
        prefix: str = "-> ",
        selected_style: int | ColorPair = 0,
        margin_height: tuple[int, int] = (0, 0),
        margin_left: int = 0,
        max_height: int = -1,
        count: Callable[[], int] | None = None,
    ) -> None:
        self._fields = fields
        self._cursor = 0
        self._selected_style = selected_style
        self._margins = margin_height
        self._max_height = max_height
        self._margin_left = margin_left
        self._prefix = prefix

        if isinstance(fields, Sequence):
            self._count_fn = lambda: len(fields)
            for _, field in fields:
                if isinstance(field, Forms):
                    field.set_prefix(self._prefix)
        else:
            self._count_fn = count if count is not None else lambda: -1

    def _get_field(self, idx: int) -> tuple[str, Callable[[], T]] | tuple[str, Forms]:
        if isinstance(self._fields, Sequence):
            return self._fields[idx]
        return self._fields(idx)

    def draw(self, stdscr: curses.window):
        """Draw menu component"""

        start, end = prepare_windowed(self._cursor, self.max_height - self._margins[1])

        for index, relative_index in enumerate(range(start, end)):
            try:
                label, content = self._get_field(relative_index)
            except (IndexError, StopIteration):
                break
            data = f"{self._prefix}{label}"
            style = 0
            if relative_index == self._cursor:
                style = curses.color_pair(int(self._selected_style))
            if hasattr(content, "display") and callable(
                getattr(content, "display", None)
            ):
                data: str = content.display()
            stdscr.addstr(self._margins[0] + index, self._margin_left, data, style)

    def get_keymap(self) -> dict[str, tuple[int, Callable[[], ReturnType]]]:
        """Get instance keymap"""
        return {
            "move_up": (self.KEYMAP_UP, self.move_up),
            "move_down": (self.KEYMAP_DOWN, self.move_down),
        }

    @property
    def max_height(self):
        """Menu max height"""
        if self._max_height == -1:
            max_height = get_terminal_size().lines
            return max_height - sum(self._margins)
        return self._max_height

    @property
    def height(self):
        """Menu height"""
        count = self._count_fn()
        if count == -1:
            return inf
        return count

    def move_down(self):
        """Move cursor down"""
        if self._cursor < self.height - 1:
            self._cursor += 1
        return ReturnType.CONTINUE

    def move_up(self):
        """Move cursor up"""
        if self._cursor > 0:
            self._cursor -= 1
        return ReturnType.CONTINUE

    def fetch(self):
        """Return callback from current cursor"""
        return self._get_field(self._cursor)

    def __repr__(self) -> str:
        return f"<{type(self).__name__} size={self.height!r}>"


class HorizontalMenu(Menu):
    """Horizontal Menu"""

    KEYMAP_LEFT = curses.KEY_LEFT
    KEYMAP_RIGHT = curses.KEY_RIGHT

    def __init__(
        self,
        fields: Fields | FieldsFn,
        prefix: str = "",
        suffix: str = "",
        selected_style: int | ColorPair = 0,
        margin_height: tuple[int, int] = (0, 0),
        margin_left: int = 0,
        max_height: int = -1,
        count: Callable[[], int] | None = None,
    ) -> None:
        super().__init__(
            fields,
            prefix,
            selected_style,
            margin_height,
            margin_left,
            max_height,
            count,
        )
        self._suffix = suffix

    def draw(self, stdscr: curses.window):
        start, end = prepare_windowed(self._cursor, visible_rows=self.max_height)
        x = 0

        for _, relative_index in enumerate(range(start, end)):
            try:
                label, content = self._get_field(relative_index)
            except (IndexError, StopIteration):
                break
            data = f"{self._prefix}{label}{self._suffix}"
            style = 0
            if relative_index == self._cursor:
                style = curses.color_pair(int(self._selected_style))
            if hasattr(content, "display") and callable(
                getattr(content, "display", None)
            ):
                data: str = content.display()
            stdscr.addstr(self._margins[0], self._margin_left + x, data, style)
            x += len(data) + 1

    def get_keymap(self) -> dict[str, tuple[int, Callable[[], ReturnType]]]:
        return {
            "move_left": (self.KEYMAP_LEFT, self.move_up),
            "move_right": (self.KEYMAP_RIGHT, self.move_down),
        }



"""Progress bar"""

# pylint: disable=no-member

import curses
import time

from .panel import Panel

PARTIAL_BLOCKS = ["", "▏", "▎", "▍", "▌", "▋", "▊", "▉", "█"]


class Progress:
    """Progress"""

    def __init__(self, smoothing: float = 0.3, min_interval: float = 0.1):
        self.smoothing = smoothing
        self.min_interval = min_interval
        self.avg_dt = None
        self.last_update = 0
        self.last_time = time.time()

    def render(
        self,
        window: curses.window | Panel,
        y: int,
        x: int,
        width: int,
        current: int,
        total: int,
        prefix: str,
    ):
        """
        Draw the progress bar at (y, x) with given width.
        Progress values are provided by the Scene.
        """

        render = window.screen if isinstance(window, Panel) else window

        now = time.time()
        # Throttle redraw
        # if now - self.last_update < self.min_interval and current < total:
        #     return

        fraction = current / max(total, 1)
        filled_cells = fraction * width
        full_blocks = int(filled_cells)
        partial_index = int((filled_cells - full_blocks) * (len(PARTIAL_BLOCKS) - 1))

        bar_str = "█" * full_blocks
        if full_blocks < width:
            bar_str += PARTIAL_BLOCKS[partial_index]
        bar_str = bar_str.ljust(width)

        # Smooth ETA
        dt = now - self.last_time
        self.last_time = now
        if self.avg_dt is None:
            self.avg_dt = dt
        else:
            self.avg_dt = (1 - self.smoothing) * self.avg_dt + self.smoothing * dt

        # if current > 0 and current < total:
        #     eta = self.avg_dt * (total - current)
        #     eta_str = f"ETA {eta:5.1f}s"
        # else:
        #     eta_str = " " * 10

        line = f"{prefix} [{bar_str}] {current}/{total}"

        # Draw in curses window
        render.addnstr(y, x, line, width + 20)  # +20 for metadata
        self.last_update = now



"""Test 2"""

# pylint: disable=no-member,no-name-in-module,wrong-import-position,missing-module-docstring
import curses
from dataclasses import dataclass
from os.path import realpath
from sys import path


p = realpath("../")
path.insert(0, p)

from lymia import Scene, on_key
from lymia.progress import Progress
from lymia.anim import Animator, move_panel
from lymia.panel import Panel
from lymia.colors import Coloring
from lymia.data import ReturnType, status
from lymia.environment import Theme
from lymia.runner import debug, run
from lymia.menu import Menu
from lymia.colors import ColorPair, color
from lymia.forms._base import Forms


class Basic(Coloring):
    """Basic colors"""

    SELECTED = ColorPair(color.BLACK, color.WHITE)


@dataclass
class Character:
    """Character"""

    name: str
    level: int
    current_hp: int
    max_hp: int
    current_mp: int
    max_mp: int
    current_energy: int
    max_energy: int


@dataclass
class DState:
    """DState"""

    character: Character
    hpbar: Progress
    mpbar: Progress
    energybar: Progress


def draw_character(screen: curses.window, state: DState):
    """Draw character HUD"""
    screen.erase()
    screen.box()
    character = state.character
    hpbar = state.hpbar
    mpbar = state.mpbar
    ebar = state.energybar
    screen.addstr(1, 2, f"{character.name} Lv. {character.level}")
    hp = character.current_hp, character.max_hp
    mp = character.current_mp, character.max_mp
    energy = character.current_energy, character.max_energy
    # screen.addstr(3, 3, f"HP: {hp[0]}/{hp[1]}")
    hpbar.render(screen, 3, 3, 7, hp[0], hp[1], "HP:")
    mpbar.render(screen, 4, 3, 7, mp[0], mp[1], "MP:")
    ebar.render(screen, 5, 3, 7, energy[0], energy[1], "E :")


def draw_action(screen: curses.window, menu: Menu):
    """Draw actions HUD"""
    screen.erase()
    screen.box()
    menu.draw(screen)


def draw_abaction(screen: curses.window, menu: Menu):
    """Draw skill/ultimate/"""
    screen.erase()
    screen.box()
    if menu is None:
        return
    menu.draw(screen)


class Root(Scene):
    """Root component"""

    render_fps = 60
    # minimal_size = (30, 120)

    def __init__(self) -> None:
        super().__init__()
        self._panels: tuple[Panel, ...]
        self._hpbar: Progress
        self._mpbar: Progress
        self._ebar: Progress
        self._character = Character("Rimu Aerisya", 100, 1, 8300, 75, 100, 185, 200)
        self._kwdargs = {
            "prefix": " ",
            "selected_style": Basic.SELECTED,
            "margin_height": (1, 2),
            "margin_left": 1,
            "max_height": 8,
        }
        self._menu = Menu(
            (
                ("Basic Attack", lambda: "Basic Attack"),
                ("Skill", lambda: "Skill"),
                ("Ultimate", lambda: "Ultimate"),
                ("Items", lambda: "Items"),
                ("Flee", lambda: "Flee"),
            ),
            **self._kwdargs,
        )
        self._ability_menu_skill = Menu(
            (
                ("Paladin's Momentum", lambda: "Skill: Paladin's Momentum"),
                ("Reformative Catalyst", lambda: "Skill: Reformative Catalyst"),
                ("Lantern of Radiance", lambda: "Skill: Lantern of Radiance"),
                ("Skill 1", lambda: "Skill: 1"),
                ("Skill 2", lambda: "Skill: 2"),
                ("Skill 3", lambda: "Skill: 3"),
                ("Skill 4", lambda: "Skill: 4"),
                ("Skill 5", lambda: "Skill: 5"),
                ("Skill 6", lambda: "Skill: 6"),
                ("Skill 7", lambda: "Skill: 7"),
                ("Skill 8", lambda: "Skill: 8"),
                ("Skill 9", lambda: "Skill: 9"),
                ("Skill 10", lambda: "Skill: 10"),
                ("Skill 11", lambda: "Skill: 11"),
                ("Skill 12", lambda: "Skill: 12"),
                ("Skill 13", lambda: "Skill: 13"),
                ("Skill 14", lambda: "Skill: 14"),
                ("Skill 15", lambda: "Skill: 15"),
                ("Skill 16", lambda: "Skill: 16"),
            ),
            **self._kwdargs,
        )
        self._ability_menu_ultimate = Menu(
            (
                ("Vainglory's Revolution", lambda: "Ultimate: Vainglory's Revolution"),
                ("The Radiance", lambda: "Ultimate: The Radiance"),
            ),
            **self._kwdargs,
        )
        self._ability_menu_items = Menu(
            (
                ("Small HP Potion", lambda: "Items: Small HP Potion"),
                ("Medium HP Potion", lambda: "Items: Medium HP Potion"),
                ("Big HP Potion", lambda: "Items: Big HP Potion"),
            ),
            **self._kwdargs,
        )
        self._abmenu = {
            "Skill": self._ability_menu_skill,
            "Ultimate": self._ability_menu_ultimate,
            "Items": self._ability_menu_items,
        }
        self._state = {"menu": self._menu}
        self._keytype = ""
        self._target_menu = 0
        self._key = 0
        self.register_keymap(self._menu)
        self._animator: Animator = Animator(self.render_fps)

    def draw(self) -> None:
        size = f"{self.height}x{self.width}"
        k = f"Key: {self._key}"
        t = f"Menu: {self._target_menu}"
        self._animator.tick()
        self.update_panels()
        status.set(
            f"FPS: {self.fps} | Screen: {size} | {k} | {t} | Action: {self._keytype}"
        )
        self.show_status()
        # self._hpbar.render(
        #     self._panels[0], 3, 8, 5, self._character.current_hp, self._character.max_hp
        # )
        if self._character.current_hp < self._character.max_hp:
            self._character.current_hp += 100
        elif self._character.current_hp >= self._character.max_hp:
            self._character.current_hp = 0

    def init(self, stdscr: curses.window):
        super().init(stdscr)
        self._hpbar = Progress()
        self._mpbar = Progress()
        self._ebar = Progress()
        self._panels = (
            Panel(
                *(8, 30, self.height - 9, 0),
                draw_character,
                DState(self._character, self._hpbar, self._mpbar, self._ebar),
            ),
            Panel(*(8, 30, self.height - 9, 30), draw_action, state=self._menu),
            Panel(*(8, 30, self.height - 9, 30), draw_abaction),
        )
        self._panels[2].hide()
        stdscr.nodelay(True)

    def handle_key(self, key: int):
        # if key == 410:
        # return ReturnType.CONTINUE
        res = super().handle_key(key)
        if key != -1:
            self._key = key
        return res

    @on_key("q")
    def quit(self):
        """quit from menu"""
        return ReturnType.EXIT

    def use_panel2(self, keytype: str):
        """Use panel 2"""
        menu = self._abmenu.get(keytype, None)
        if menu is None:
            return
        self._panels[2].set_state(menu)
        anim = move_panel(
            self._panels[1], 30, self.height - 9, 60, self.height - 9, 0.5
        )
        anim.on_complete(self.panel2_oncomplete)
        self._animator.add(anim)

    def revert(self):
        """Revert to Panel 1"""
        self._panels[2].hide()
        self._panels[2].set_state(None)
        anim = move_panel(
            self._panels[1], 60, self.height - 9, 30, self.height - 9, 0.5
        )
        anim.on_complete(self.panel1_oncomplete)
        self._animator.add(anim)

    def panel1_oncomplete(self, _: Panel):
        """Panel 1 on complete"""
        self.register_keymap(self._menu)
        self._target_menu = 0

    def panel2_oncomplete(self, _: Panel):
        """Panel 2 on complete"""
        self._panels[2].show()
        menu: Menu = self._panels[2].get_state()  # type: ignore
        self.register_keymap(menu)
        self._target_menu = 1

    @on_key(curses.KEY_RIGHT)
    def select(self):
        """Select from skill menu"""
        if not self._animator.is_empty:
            return ReturnType.CONTINUE
        _, ability = (
            self._menu.fetch()
            if self._target_menu == 0
            else self._panels[2].get_state().fetch()  # type: ignore
        )
        if not isinstance(ability, Forms) and self._panels[2].panel.hidden():
            keytype = ability()
            if keytype is None:
                return ReturnType.CONTINUE
            if keytype == "Flee":
                return ReturnType.EXIT
            if keytype in ("Basic Attack",):
                self._keytype = keytype
                return ReturnType.CONTINUE
            self.use_panel2(keytype)
        if not isinstance(ability, Forms) and self._panels[2].panel.hidden() is False:
            keytype = ability()
            if keytype is None:
                return ReturnType.CONTINUE
            self._keytype = keytype
            self.revert()

        return ReturnType.CONTINUE

    @on_key(curses.KEY_LEFT)
    def unselect(self):
        """Unselect"""
        if self._panels[2].panel.hidden() is False:
            self.revert()
        return ReturnType.CONTINUE


@debug
def init():
    """init"""
    root = Root()
    env = Theme(0, Basic())
    return root, env


if __name__ == "__main__":
    run(init)



# pylint: disable=no-member,no-name-in-module,wrong-import-position,missing-module-docstring
import curses
from os.path import realpath
from sys import path

from typing import Callable


p = realpath("../")
print(p)
path.insert(0, p)

from lymia import const, run, Scene, on_key, Menu, MenuFormScene, Forms, HorizontalMenu
from lymia.data import SceneResult, ReturnType, status
from lymia.colors import Coloring, ColorPair, color
from lymia.environment import Theme
from lymia.forms import FormFields, Text, Password
from lymia.runner import debug

class Basic(Coloring):
    """Basic color pair"""

    SELECTED = ColorPair(color.BLACK, color.YELLOW)


class MinuteClock(Scene):
    """Set an alarm (minute)"""

    def generate_time_minute(self, index: int):
        """Generate time hour by index"""
        return (
            str(index % 60),
            lambda: (
                status.set(f"Clock set: {self._time:0>2}:{index % 60:0>2}"),
                ReturnType.CONTINUE,
            )[1],
        )

    def __init__(self, time: int) -> None:
        self.margin_top = 2
        self._time = time
        super().__init__()
        self._menu: Menu[int] = Menu(
            self.generate_time_minute,
            prefix="   ",
            margin_height=(self.margin_top, 2),
            selected_style=Basic.SELECTED,
        )
        self.register_keymap(self._menu)

    def draw(self):
        self.screen.addstr(0, 0, f"Minutes (Hour selected: {self._time})")
        self._menu.draw(self.screen)
        self.show_status()

    @on_key(curses.KEY_LEFT)
    def back(self):
        """Back"""
        return ReturnType.BACK

    @on_key(curses.KEY_RIGHT)
    def exec_menu(self):
        """Pops menu and returns component"""
        comp: Callable[[], ReturnType] = self._menu.fetch()[1]  # type: ignore
        if not isinstance(comp, Forms):
            return comp()
        return ReturnType.CONTINUE

    @on_key("q")
    def quit(self):
        """On quit"""
        return ReturnType.EXIT


class Clock(Scene):
    """Set an alarm"""

    def generate_time_hour(self, index: int):
        """Generate time hour by index"""
        return str(index % 24), lambda: self.to_minute_clock(index % 24)

    def to_minute_clock(self, time: int):
        """To minute clock"""
        return MinuteClock(time)

    def __init__(self) -> None:
        self.margin_top = 2
        super().__init__()
        self._menu: Menu[MinuteClock] = Menu(
            self.generate_time_hour,
            prefix="   ",
            margin_height=(self.margin_top, 4),
            selected_style=Basic.SELECTED,
        )
        self.register_keymap(self._menu)

    def draw(self):
        self.screen.addstr(0, 0, "Hours")
        self._menu.draw(self.screen)
        self.show_status()

    @on_key(curses.KEY_LEFT)
    def back(self):
        """Back"""
        return ReturnType.BACK

    @on_key(curses.KEY_RIGHT)
    def exec_menu(self):
        """Pops menu and returns component"""
        comp: Callable[[], Scene] = self._menu.fetch()[1]  # type: ignore
        if not isinstance(comp, Forms):
            return SceneResult(comp())
        return ReturnType.CONTINUE

    @on_key("q")
    def quit(self):
        """On quit"""
        return ReturnType.EXIT


class Settings(MenuFormScene):
    """Settings component"""

    def __init__(self) -> None:
        self.margin_top = 2
        fields = FormFields(
            (Text("Username"), Password("Password")),
            self.margin_top,
        )

        menu: Menu = Menu(
            fields.to_menu_fields(),
            prefix="   ",
            margin_height=(self.margin_top, 2),
            selected_style=Basic.SELECTED,
        )
        super().__init__(menu)

    def draw(self):
        self.screen.addstr(0, 0, "App settings")
        super().draw()
        self.show_status()

    @on_key(curses.KEY_LEFT)
    def back(self):
        """Back"""
        return ReturnType.BACK

    @on_key(curses.KEY_RIGHT)
    def enter_mode(self):
        """Enter edit mode"""
        return self.select_menu_item()


class Settings2(Scene):
    """Settings 2"""

    def __init__(self):
        super().__init__()
        self.margin_top = 2
        fields = (
            ("Button 1", lambda: status.set("Button 1 clicked")),
            ("Button 2", lambda: status.set("Button 2 clicked")),
        )
        self._menu = HorizontalMenu(
            fields,
            prefix="",
            selected_style=Basic.SELECTED,
            margin_height=(self.margin_top, 0),
            margin_left=0,
        )
        self.register_keymap(self._menu)

    def draw(self):
        self.screen.addstr(0, 0, "App settings")
        self._menu.draw(self.screen)
        self.show_status()

    @on_key("q")
    def back(self):
        """Back"""
        return ReturnType.BACK

    @on_key(const.KEY_ESC)
    def quit(self):
        """Quit"""
        return ReturnType.EXIT

    @on_key(curses.KEY_ENTER, const.KEY_ENTER)
    def enter_mode(self):
        """Enter edit mode"""
        _, callback = self._menu.fetch()
        if not isinstance(callback, Forms):
            callback()
        return ReturnType.CONTINUE

class Root(Scene):
    """Root component"""

    def __init__(self) -> None:
        super().__init__()
        self._menu: Menu[ReturnType | Scene] = Menu(
            (
                ("Settings", lambda: Settings()),  # pylint: disable=unnecessary-lambda
                ("Settings", lambda: Settings2()),  # pylint: disable=unnecessary-lambda
                ("Clock", lambda: Clock()),  # pylint: disable=unnecessary-lambda
            ),
            margin_height=(2, 2),
            selected_style=Basic.SELECTED,
        )
        self.register_keymap(self._menu)

    def draw(self) -> None | ReturnType:
        self.screen.addstr(0, 0, "Hello, World!")
        self.show_status()
        self._menu.draw(self.screen)

    @on_key(curses.KEY_RIGHT)
    def exec_menu(self):
        """Pops menu and returns component"""
        comp: Callable[[], Scene] = self._menu.fetch()[1]  # type: ignore
        if not isinstance(comp, Forms):
            return SceneResult(comp())
        return ReturnType.CONTINUE

    @on_key("q")
    def quit(self):
        """On quit"""
        return ReturnType.EXIT


@debug
def init():
    """main function"""
    root = Root()
    env = Theme(const.CURSOR_HIDDEN, Basic())
    return root, env


if __name__ == "__main__":
    run(init)



"""Curses utility"""

# pylint: disable=no-member

# from os import get_terminal_size
import curses
from contextlib import contextmanager
from typing import TypeVar

T = TypeVar("T")

@contextmanager
def hide_system(stdscr: curses.window):
    """Hide the app UI to parent TTY"""
    curses.endwin()
    try:
        yield
    finally:
        stdscr.refresh()
        curses.doupdate()

def clear_line(stdscr: curses.window, line: int):
    """Clear a line"""
    stdscr.addstr(line, 0, " " * (stdscr.getmaxyx()[1] - 1))

@contextmanager
def clear_line_yield(stdscr: curses.window, line: int):
    """Clear line through context var"""
    clear_line(stdscr, line)
    try:
        yield
    finally:
        clear_line(stdscr, line)
        stdscr.refresh()

def windowed(data: list[T] | tuple[T, ...], start: int, end: int):
    """Return an enumerated, windowed list based on start and end.
    Start/end must incorporated values returned by prepare_windowed"""
    return list(enumerate(data))[start:end]


def prepare_windowed(index: int, visible_rows: int):
    """Return min/max length for windowed function"""
    minln = max(0, index - (visible_rows // 2))
    maxln = (
        visible_rows if index <= (visible_rows // 2) else (index + visible_rows // 2)
    )
    return minln, maxln



"""Password form field"""

from ._base import Forms


class Password(Forms):
    """Password form fields"""

    def __init__(
        self,
        label: str,
        value: str = "",
        suffix: str = ": ",
        margin_left: int = 0,
        field_pos: int = 0,
        style: int = 0,
    ) -> None:
        super().__init__(label, value, suffix, margin_left, field_pos, style)

    @property
    def displayed_value(self):
        return "*" * (len(self._buffer))



"""Lymia"""

from .progress import Progress
from .panel import Panel
from .anim import Animation, Animator, KeyframeAnimation
from .runner import bootstrap, run
from .menu import Menu, HorizontalMenu
from .scene import Scene, MenuFormScene, on_key
from .data import status, ReturnInfo, ReturnType
from .utils import hide_system, clear_line, clear_line_yield
from .forms import Password, Text, FormFields, Forms

__version__ = "0.0.3"
__all__ = [
    'bootstrap',
    'run',
    'Scene',
    'on_key',
    'status',
    "status",
    "ReturnInfo",
    "ReturnType",
    "hide_system",
    "clear_line",
    "clear_line_yield",
    "Menu",
    "MenuFormScene",
    "Password",
    "Text",
    'FormFields',
    "Forms",
    "Animator",
    "Animation",
    'KeyframeAnimation',
    'Panel',
    'Progress',
    'HorizontalMenu'
]



"""Generic runners"""

# pylint: disable=no-member,no-name-in-module

import curses
import curses.panel
from functools import wraps
from os import get_terminal_size
from time import sleep, perf_counter
from typing import Callable, ParamSpec

from .environment import Theme
from .data import SceneResult, ReturnType
from .scene import Scene
from .utils import hide_system

Ps = ParamSpec("Ps")
DEBUG = True


def _oob_check(sizes: tuple[int, int], minsize: tuple[int, int]):
    """Check whether current sizes is enough for minsize"""
    return not (sizes[0] < minsize[0] or sizes[1] < minsize[1])


def _oob_runner(render: curses.window, fps: int, minsize: tuple[int, int]):
    sizes = f"{minsize[0]}x{minsize[1]}"
    with hide_system(render):
        while True:
            size: tuple[int, int] = get_terminal_size()[::-1]  # type: ignore
            csize = f"{size[0]}x{size[1]}"
            print(f"Cannot go below {sizes} ({csize})", end="\r")
            if _oob_check(size, minsize):
                return
            sleep(1 / fps)


def check_resize(
    render: curses.window,
    scene: Scene,
    sizes: tuple[int, int],
    minsize: tuple[int, int],
):
    """Check scene resize"""
    new_size = render.getmaxyx()
    if sizes != new_size and scene.auto_resize:
        # scene.init(render)
        sizes = new_size
        minsize = max(scene.minimal_size, minsize)
        if sizes[0] < minsize[0] or sizes[1] < minsize[1]:
            _oob_runner(render, scene.render_fps, minsize)
        return new_size
    return sizes


def wait(render_fps: int, start: float):
    """Wait for FPS"""
    if render_fps > 1:  # why do we add this anyway?
        wend = perf_counter()
        wdt = wend - start
        frametime = 1 / render_fps
        remaining = frametime - wdt
        if remaining > 0:
            sleep(remaining)


def on_back(ret: ReturnType | None, stack: list[Scene], current: Scene):
    """On back event"""
    if ret in (ReturnType.BACK, ReturnType.ERR_BACK):
        stack.pop()
        current.on_unmount()
        return True
    return False


def screen_update():
    """Update screne"""
    curses.panel.update_panels()
    curses.doupdate()


def report_fps(scene: Scene, fc: int, end: float, start: float):
    """Report current FPS to current scene"""
    if end - start >= 1.0:
        scene.fps = fc / (end - start)
        fc = 0
        start = end
    return fc, start


def draw(scene: Scene, delta: float, start: float):
    """Draw current scene"""
    if scene.animator:
        scene.animator.deltatime = delta
    ret = scene.draw()
    screen_update()
    wait(scene.render_fps, start)
    return ret


def on_exit(ret: ReturnType, stack: list[Scene]):
    """On exit"""
    if ret == ReturnType.EXIT:
        for scene in stack:
            scene.on_unmount()
        stack.clear()
        return True
    return False


def on_push(result: SceneResult, screen: curses.window, stack: list[Scene]):
    """On push"""
    if isinstance(result, SceneResult):
        scene = result.scene
        scene.init(result.target or screen)
        stack.append(scene)
        return result.target or screen
    return False


def process_scene_result(
    result: ReturnType | SceneResult,
    stack: list[Scene],
    current: Scene,
    screen: curses.window,
    root_screen: curses.window,
):
    """Process the result of handle_key and update stack/render as needed."""
    if result == ReturnType.RETURN_TO_MAIN:
        while len(stack) != 1:
            scene = stack.pop()
            scene.on_unmount()
        return screen
    if on_exit(result, stack):  # type: ignore
        return False
    if on_back(result, stack, current):  # type: ignore
        return screen
    pushed = on_push(result, root_screen, stack)  # type: ignore
    if pushed:
        return pushed
    return None


def runner(stdscr: curses.window, root: Scene, env: Theme | None = None):
    """Run the whole scheme"""
    stack: list[Scene] = [root]
    render = stdscr
    root.init(render)
    sizes = render.getmaxyx()
    root_minsize = root.minimal_size
    delta: float = 0
    start = end = 0
    frame_count = 0
    window_start = perf_counter()

    if env:
        env.apply()

    while stack:
        start = perf_counter()
        current = stack[-1]

        if current.should_clear:
            render.erase()

        sizes = check_resize(render, current, sizes, root_minsize)
        ret = draw(current, delta, start)

        if on_back(ret, stack, current):
            render = stdscr
            continue

        key = render.getch()
        result = current.handle_key(key)
        screen_update()

        # comp.handle_key() may mutate display
        end = perf_counter()
        delta = end - start
        frame_count += 1
        frame_count, window_start = report_fps(current, frame_count, end, window_start)

        new_render = process_scene_result(result, stack, current, render, stdscr)
        if new_render is False:
            break
        if new_render is None:
            continue
        render = new_render


def bootstrap(fn: Callable[Ps, tuple[Scene, Theme | None]]):
    """Run the app, must be used as decorator like:

    @bootstrap
    def init():
        ...
    """

    @wraps(fn)
    def inner(*args, **kwargs):
        return curses.wrapper(runner, *fn(*args, **kwargs))

    return inner


def debug(fn: Callable[Ps, tuple[Scene, Theme | None]]):
    """Enable debug mode"""
    global DEBUG  # pylint: disable=global-statement
    DEBUG = True
    return fn


def run(_fn: Callable[Ps, tuple[Scene, Theme | None]], *args, **kwargs):
    """Run main function, the structure must be similiar of `@bootstrap` target function."""
    return curses.wrapper(runner, *_fn(*args, **kwargs))



"""App environment"""
# pylint: disable=no-member

import curses

from lymia.colors import Coloring

class Theme:
    """App theme"""
    def __init__(self, cursor_style: int, style: Coloring) -> None:
        self._style = style
        self._cursor_style = cursor_style

    def apply(self):
        """Apply current theme"""
        if self._style:
            curses.start_color()
            for style in self._style:
                curses.init_pair(int(style), style.fg, style.bg)
        curses.curs_set(self._cursor_style)



"""Animation"""

from typing import Callable
from .panel import Panel


class Animator:
    """Animator"""

    def __init__(self, fps: int):
        if fps <= 0:
            raise ValueError(f"Invalid FPS target: {fps}")
        self._fps = fps
        self._dt = 1 / fps
        self._animations: "list[BaseAnim]" = []

    def add(self, animation: "BaseAnim"):
        """Add animations"""
        self._animations.append(animation)

    @property
    def is_empty(self):
        """Is empty?"""
        return len(self._animations) == 0

    @property
    def deltatime(self):
        """Delta time"""
        return self._dt

    @deltatime.setter
    def deltatime(self, dt: float):
        """delta time"""
        self._dt = dt

    def tick(self):
        """Tick"""
        dt = self.deltatime

        for anim in self._animations[:]:
            anim.update(dt)
            if anim.finished:
                self._animations.remove(anim)


class BaseAnim:
    """Base class for all animation"""

    def __init__(self) -> None:
        self._finished = False
        self._target: Panel
        self._callback: Callable[[Panel], None] = lambda panel: panel.show()

    def on_complete(self, callback: Callable[[Panel], None]):
        """On complete callback"""
        self._callback = callback

    @property
    def finished(self):
        """Is finished"""
        return self._finished

    def update(self, dt: float):
        """Update"""
        raise NotImplementedError


class Animation(BaseAnim):
    """Animation handler"""

    def __init__(
        self,
        target: Panel,
        duration: int | float,
        update_fn: Callable[[Panel, float], None],
    ):
        super().__init__()
        self._target = target
        self._duration = duration
        self._update_fn = update_fn
        self._elapsed = 0

    def update(self, dt: int):
        """Update"""
        if self._finished:
            return
        self._elapsed += dt
        t = min(self._elapsed / self._duration, 1.0)  # normalized progress
        self._update_fn(self._target, t)
        if self._elapsed >= self._duration:
            self._finished = True
            self._callback(self._target)


class KeyframeAnimation(BaseAnim):
    """Keyframe animation"""

    def __init__(
        self,
        target: Panel,
        keyframes: list[tuple[int, tuple[int, int]]],
        duration: int | float,
        interp_fn: Callable[[int, int, float], float] | None = None,
    ):
        super().__init__()
        self._target = target
        self._keyframes = sorted(keyframes, key=lambda k: k[0])
        self._duration = duration
        self._elapsed = 0
        self._interp_fn = interp_fn or self.linear

    def linear(self, a, b, t):
        """Linear changes"""
        return a + (b - a) * t

    def update(self, dt: float):
        """Update"""
        if self.finished:
            return
        self._elapsed += dt
        progress = min(self._elapsed / self._duration, 1.0)

        # find current keyframe segment
        for i in range(len(self._keyframes) - 1):
            t1, v1 = self._keyframes[i]
            t2, v2 = self._keyframes[i + 1]
            if t1 <= progress <= t2:
                local_t = (progress - t1) / (t2 - t1)
                x = self._interp_fn(v1[0], v2[0], local_t)
                y = self._interp_fn(v1[1], v2[1], local_t)
                self._target.move(int(y), int(x))
                break

        if progress >= 1.0:
            self._finished = True
            self._callback(self._target)


def move_panel(panel: Panel, x1: int, y1: int, x2: int, y2: int, duration: int | float):
    """Move panels around"""

    def updater(target: Panel, t: float):
        x = int(x1 + (x2 - x1) * t)
        y = int(y1 + (y2 - y1) * t)
        target.move(x, y)

    return Animation(panel, duration, updater)



"""Form fields"""

from ._base import Forms
from .text import Text
from .password import Password


class FormFields:
    """Form fields"""

    def __init__(self, fields: tuple[Forms, ...], margin_top: int = 0) -> None:
        self._fields = fields
        for index, field in enumerate(fields):
            field.set_field_pos(margin_top + index)

    def to_menu_fields(self):
        """To menu fields"""
        return tuple((field.label, field) for field in self._fields)


__all__ = ["FormFields", "Forms", "Text", "Password"]



"""Colors"""

from . import _color_const as color

class ColorPair:
    """Define colors with fg and bg"""
    def __init__(self, fg: int, bg: int) -> None:
        self._fg = fg
        self._bg = bg
        self._id = -1

    @property
    def id(self):
        """Id"""
        return self._id

    @property
    def fg(self):
        """Foreground"""
        return self._fg

    @property
    def bg(self):
        """background"""
        return self._bg

    def __int__(self):
        if self._id == -1:
            raise ValueError("This color must be initialized")
        return self._id

    def set_id(self, cid: int):
        """Set this color id"""
        self._id = cid

    def __repr__(self) -> str:
        return f"<ColorPair[{self._id}] fg={self._fg} bg={self._bg}>"

class Coloring:
    """Coloring"""
    def __init__(self) -> None:
        self._fields: dict[str, ColorPair] = {}
        for index, (key, value) in enumerate(type(self).__dict__.items()):
            if not isinstance(value, ColorPair):
                continue
            value.set_id(index)
            self._fields[key] = value

    def __iter__(self):
        return (field for field in self._fields.values())

    def __bool__(self):
        return bool(self._fields)

__all__ = ['ColorPair', 'color', "Coloring"]
